/* as.c - GAS main program.
   Copyright (C) 1987 Free Software Foundation, Inc.

This file is part of GAS, the GNU Assembler.

GAS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GAS; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * Main program for AS; a 32-bit assembler of GNU.
 * Understands command arguments.
 * Has a few routines that don't fit in other modules because they
 * are shared.
 *
 *
 *			bugs
 *
 * : initialisers
 *	Since no-one else says they will support them in future: I
 * don't support them now.
 *
 */

#include <signal.h>

#define COMMON
#include "as.h"
#include "struc-symbol.h"
#include "write.h"
		/* Warning!  This may have some slightly strange side effects
		   if you try to compile two or more assemblers in the same
		   directory!
		 */

#ifndef SIGTY
#define SIGTY int
#endif

SIGTY got_sig();

#ifdef DONTDEF
static char * gdb_symbol_file_name;
long int gdb_begin();
#endif

char *myname;		/* argv[0] */
extern char version_string[];
#ifdef NeXT
extern char next_version[];
extern int arch_multiple;
#endif NeXT

#ifdef NeXT	/* .include feature */
/* DJA
 * .include "file" looks in source file dir, then stack.
 * -I directories are added to the end, then the defaults are added.
 */
struct directory_stack include_defaults[] =
  {
    { 0, "/NextDeveloper/Headers/" },
    { 0, "/LocalDeveloper/Headers/" },
    { 0, NULL }
  };

struct directory_stack *include = 0;		/* First dir to search */
struct directory_stack *include_tail = 0;	/* Last in chain */
#endif NeXT	/* .include feature */

#ifdef NeXT
int force_cpusubtype_ALL = 0;
cpu_subtype_t archflag_cpusubtype = -1;
cpu_subtype_t cpusubtype = 
#ifdef M68K
	    CPU_SUBTYPE_MC680x0_ALL;
#endif
#ifdef I860
	    CPU_SUBTYPE_I860_ALL;
#endif
#ifdef M88K
	    CPU_SUBTYPE_MC88000_ALL;
#endif
#ifdef I386
	    CPU_SUBTYPE_I386_ALL;
#endif
#ifdef M98K
	    CPU_SUBTYPE_MC98000_ALL;
#endif
#endif /* NeXT */


main(argc,argv)
int	argc;
char	**argv;
{
	int	work_argc;	/* variable copy of argc */
	char	**work_argv;	/* variable copy of argv */
	char	*arg;		/* an arg to program */
	char	a;		/* an arg flag (after -) */
	static int sig[] = { SIGHUP, SIGINT, SIGPIPE, SIGTERM, 0};

	extern int bad_error;	/* Did we hit a bad error ? */

	char	*stralloc();	/* Make a (safe) copy of a string. */
	void	symbol_begin();
	void	read_begin();
	void	write_object_file();
#ifdef NeXT
	char *specific_archflag = NULL;
#endif

	for(a=0;sig[a]!=0;a++)
		if(signal(sig[a], SIG_IGN) != SIG_IGN)
			signal(sig[a], got_sig);

	myname=argv[0];
	bzero (flagseen, sizeof(flagseen)); /* aint seen nothing yet */
	out_file_name	= "a.out";	/* default .o file */
	symbol_begin();		/* symbols.c */
	subsegs_begin();		/* subsegs.c */
	read_begin();			/* read.c */
	md_begin();			/* MACHINE.c */
	input_scrub_begin();		/* input_scrub.c */
#ifdef DONTDEF
	gdb_symbol_file_name = 0;
#endif
	/*
	 * Parse arguments, but we are only interested in flags.
	 * When we find a flag, we process it then make it's argv[] NULL.
	 * This helps any future argv[] scanners avoid what we processed.
	 * Since it is easy to do here we interpret the special arg "-"
	 * to mean "use stdin" and we set that argv[] pointing to "".
	 * After we have munged argv[], the only things left are source file
	 * name(s) and ""(s) denoting stdin. These file names are used
	 * (perhaps more than once) later.
	 */
	work_argc = argc-1;		/* don't count argv[0] */
	work_argv = argv+1;		/* skip argv[0] */
	for (;work_argc--;work_argv++) {
		arg = * work_argv;	/* work_argv points to this argument */

		if (*arg!='-')		/* Filename. We need it later. */
			continue;	/* Keep scanning args looking for flags. */
		if (arg[1] == '-' && arg[2] == 0) {
			/* "--" as an argument means read STDIN */
			/* on this scan, we don't want to think about filenames */
			* work_argv = "";	/* Code that means 'use stdin'. */
			continue;
		}
				/* This better be a switch. */
		arg ++;		/* -> letter. */

		while (a = * arg)  {/* scan all the 1-char flags */
			arg ++;	/* arg -> after letter. */
			a &= 0x7F;	/* ascii only please */
#ifdef NeXT
			if (flagseen[a] && (a!='I') && (a!='a') && (a!='f'))
#else
			if (flagseen[a])
#endif
				as_warn("%s: Flag option -%c has already been seen!",myname,a);
#ifdef NeXT
			if(a != 'f')
#endif
			    flagseen[a] = TRUE;
			switch (a) {
			case 'f':
#ifdef NeXT
				if(strcmp(arg-1, "force_cpusubtype_ALL") == 0){
				    force_cpusubtype_ALL = 1;
				    arg = "";	/* Finished with this arg. */
				    break;
				}
				flagseen[a] = TRUE;
#endif /* NeXT */
				break;	/* -f means fast - no need for "app" preprocessor. */

#ifndef NeXT
			/* NeXT, since this does nothing anywhere in the code
			   that is used by NeXT this flag is not recognized */
			case 'D':
				/* DEBUG is implemented: it debugs different */
				/* things to other people's assemblers. */
				break;
#endif NeXT

#ifdef DONTDEF
			case 'G':	/* GNU AS switch: include gdbsyms. */
				if (*arg)	/* Rest of argument is file-name. */
					gdb_symbol_file_name = stralloc (arg);
				else if (work_argc) {	/* Next argument is file-name. */
					work_argc --;
					* work_argv = NULL; /* Not a source file-name. */
					gdb_symbol_file_name = * ++ work_argv;
				} else
					as_warn( "%s: I expected a filename after -G",myname);
				arg = "";	/* Finished with this arg. */
				break;
#endif

#ifndef WORKING_DOT_WORD
			case 'k':
				break;
#endif

			case 'L': /* -L means keep L* symbols */
				break;

			case 'o':
				if (*arg)	/* Rest of argument is object file-name. */
					out_file_name = stralloc (arg);
				else if (work_argc) {	/* Want next arg for a file-name. */
					* work_argv = NULL; /* This is not a file-name. */
					work_argc--;
					out_file_name = * ++ work_argv;
				} else
					as_warn("%s: I expected a filename after -o. \"%s\" assumed.",myname,out_file_name);
				arg = "";	/* Finished with this arg. */
				break;

			case 'R':
				/* -R means put data into text segment */
#ifdef NeXT
				as_warn("%s: -R option not supported (used the "
					".const directive)", myname);
				flagseen['R'] = FALSE;
#endif NeXT
				break;

			case 'v':
#ifdef	VMS
				{
				extern char *compiler_version_string;
				compiler_version_string = arg;
				}
#else /* not VMS */
#ifdef NeXT
				fprintf(stderr,"NeXT Computer, Inc. version "
					"%s, ", next_version);
#endif NeXT
				fprintf(stderr,version_string);
				if(*arg && strcmp(arg,"ersion"))
					as_warn("Unknown -v option ignored");
#endif
				while(*arg) arg++;	/* Skip the rest */
				break;

			case 'W':
				/* -W means don't warn about things */
				break;

#ifdef NeXT	/* .include feature */
			case 'I': /* DJA -- Add directory to path for includes*/
				{
				  struct directory_stack *dirtmp;

				  dirtmp = (struct directory_stack *)
				    xmalloc (sizeof (struct directory_stack));
				  /* New one goes on the end */
				  dirtmp->next = 0;
				  if (include == 0)
				    include = dirtmp;
				  else
				    include_tail->next = dirtmp;
				  /* Tail follows the last one */
				  include_tail = dirtmp;
				  /* Rest of argument is include file-name. */
				  if (*arg)
				    dirtmp->fname = stralloc (arg);
				  else if (work_argc)
				    {	/* Want next arg for a file-name. */
				      /* This is not a file-name. */
				      * work_argv = NULL;
				      work_argc--;
				      dirtmp->fname = * ++ work_argv;
				    }
				  else
				    as_warn("I expected a filename after -I.");
				  arg = "";	/* Finished with this arg. */
				}
				break;
#endif NeXT	/* .include feature */

#ifdef NeXT	/* generate stabs for debugging assembly code */
			case 'g':
				break;
#endif NeXT	/* generate stabs for debugging assembly code */

#ifdef NeXT	/* mutiple arch stuff, see driver.c */
			case 'V':
				/* as driver's -V, verbose, flag */
				break;
			case 'a':
				if(strcmp(arg-1, "arch_multiple") == 0){
				    arch_multiple = 1;
				    arg = "";	/* Finished with this arg. */
				    break;
				}
				else if(strcmp(arg-1, "arch") == 0){
				    arg = "";	/* Finished with this arg. */
				    /* Want next arg for a <arch_type> */
				    if(work_argc){
					/* This, "-arch" is not a file-name. */
					*work_argv = NULL;
					work_argc--;
					work_argv++;
#ifdef M68K
					if(strcmp(*work_argv, "m68030") == 0){
					    if(archflag_cpusubtype != -1 &&
					       archflag_cpusubtype !=
						CPU_SUBTYPE_MC68030_ONLY)
						as_fatal("can't specify both "
						    "-arch m68030 and -arch "
						    "m68040");
					    specific_archflag = *work_argv;
					    archflag_cpusubtype =
						CPU_SUBTYPE_MC68030_ONLY;
					}
					else if(strcmp(*work_argv,
							"m68040") == 0){
					    if(archflag_cpusubtype != -1 &&
					       archflag_cpusubtype !=
						CPU_SUBTYPE_MC68040)
						as_fatal("can't specify both "
						    "-arch m68030 and -arch "
						    "m68040");
					    specific_archflag = *work_argv;
					    archflag_cpusubtype =
						CPU_SUBTYPE_MC68040;
					}
					else if(strcmp(*work_argv, "m68k") != 0)
					    as_fatal("I expected 'm68k', "
						"'m68030' or 'm68040' after "
						"-arch for this assembler.");
#endif
#ifdef M88K
					if(strcmp(*work_argv, "m88k") != 0)
					    as_fatal("I expected 'm88k' after "
						   "-arch for this assembler.");
#endif
#ifdef I860
					if(strcmp(*work_argv, "i860") != 0)
					    as_fatal("I expected 'i860' after "
						   "-arch for this assembler.");
#endif
#ifdef I386
					if(strcmp(*work_argv, "i486") == 0){
					    if(archflag_cpusubtype != -1 &&
					       archflag_cpusubtype !=
						    CPU_SUBTYPE_486)
						as_fatal("can't specify more "
						   "than one -arch flag ");
					    specific_archflag = *work_argv;
					    archflag_cpusubtype =
						CPU_SUBTYPE_486;
					}
					else if(strcmp(*work_argv,
						       "i486SX") == 0){
					    if(archflag_cpusubtype != -1 &&
					       archflag_cpusubtype !=
						    CPU_SUBTYPE_486SX)
						as_fatal("can't specify more "
						   "than one -arch flag ");
					    specific_archflag = *work_argv;
					    archflag_cpusubtype =
						CPU_SUBTYPE_486SX;
					}
					else if(strcmp(*work_argv, "i586") ==0){
					    if(archflag_cpusubtype != -1 &&
					       archflag_cpusubtype !=
						    CPU_SUBTYPE_586)
						as_fatal("can't specify more "
						   "than one -arch flag ");
					    specific_archflag = *work_argv;
					    archflag_cpusubtype =
						CPU_SUBTYPE_586;
					}
					else if(strcmp(*work_argv,
						       "i586SX") == 0){
					    if(archflag_cpusubtype != -1 &&
					       archflag_cpusubtype !=
						    CPU_SUBTYPE_586SX)
						as_fatal("can't specify more "
						   "than one -arch flag ");
					    specific_archflag = *work_argv;
					    archflag_cpusubtype =
						CPU_SUBTYPE_586SX;
					}
					else if(strcmp(*work_argv, "i386") != 0)
					    as_fatal("I expected 'i386', 'i486'"
					       " 'i486SX', 'i586' or 'i586SX' "
					       "after -arch for this "
					       "assembler.");
#endif
				    }
				    else
					as_fatal("I expected an <arch_type> "
						 "after -arch.");
				    break;
				}
				/* fall through for non -arch flag */
#endif NeXT	/* mutiple arch stuff, see driver.c */
			default:
				--arg;
				if(md_parse_option(&arg,&work_argc,&work_argv)==0)
					as_warn("%s: I don't understand '%c' flag!",myname,a);
				if(arg && *arg)
					arg++;
				break;
			}
		}
		/*
		 * We have just processed a "-..." arg, which was not a
		 * file-name. Smash it so the
		 * things that look for filenames won't ever see it.
		 *
		 * Whatever work_argv points to, it has already been used
		 * as part of a flag, so DON'T re-use it as a filename.
		 */
		*work_argv = NULL; /* NULL means 'not a file-name' */
	}
#ifdef NeXT
	if(force_cpusubtype_ALL && specific_archflag)
	    archflag_cpusubtype = -1;
#endif
#ifdef DONTDEF
	if (gdb_begin(gdb_symbol_file_name) == 0)
		flagseen ['G'] = 0;	/* Don't do any gdbsym stuff. */
#endif
	/* Here with flags set up in flagseen[]. */
	perform_an_assembly_pass(argc,argv); /* Assemble it. */
	if (seen_at_least_1_file() && !bad_error)
		write_object_file();/* relax() addresses then emit object file */
	input_scrub_end();
	md_end();			/* MACHINE.c */
#ifndef	VMS
	exit(bad_error);			/* WIN */
#else	/* VMS */
	exit(!bad_error);			/* WIN */
#endif	/* VMS */
}
 

/*			perform_an_assembly_pass()
 *
 * Here to attempt 1 pass over each input file.
 * We scan argv[*] looking for filenames or exactly "" which is
 * shorthand for stdin. Any argv that is NULL is not a file-name.
 * We set need_pass_2 TRUE if, after this, we still have unresolved
 * expressions of the form (unknown value)+-(unknown value).
 *
 * Note the un*x semantics: there is only 1 logical input file, but it
 * may be a catenation of many 'physical' input files.
 */
perform_an_assembly_pass (argc, argv)
int	argc;
char **	argv;
{
	char *	buffer;		/* Where each bufferful of lines will start. */
	void	read_a_source_file();
	int saw_a_file = 0;

	text_fix_root		= NULL;
	data_fix_root		= NULL;
	need_pass_2		= FALSE;

	argv++;			/* skip argv[0] */
	argc--;			/* skip argv[0] */
	while (argc--) {
		if (*argv) {		/* Is it a file-name argument? */
			/* argv -> "" if stdin desired, else -> filename */
			if (buffer = input_scrub_new_file (*argv) ) {
				saw_a_file++;
				read_a_source_file(buffer);
			}
		}
		argv++;			/* completed that argv */
	}
	if(!saw_a_file)
		if(buffer = input_scrub_new_file("") )
			read_a_source_file(buffer);
}

/*
 *			stralloc()
 *
 * Allocate memory for a new copy of a string. Copy the string.
 * Return the address of the new string. Die if there is any error.
 */

char *
stralloc (str)
char *	str;
{
	register char *	retval;
	register long int	len;

	len = strlen (str) + 1;
	retval = xmalloc (len);
	(void)strcpy (retval, str);
	return (retval);
}

lose()
{
	as_fatal( "%s: 2nd pass not implemented - get your code from random(3)",myname );
}

SIGTY
got_sig(sig)
int sig;
{
	static here_before = 0;

	as_bad("Interrupted by signal %d",sig);
	if(here_before++)
		exit(1);
}

/* end: as.c */
