#import "Block.h"


static id <BlockGetter> blockGetter;

@implementation Block

+(void)setBlockGetter: (id <BlockGetter>)theGetter { blockGetter = theGetter; }
-(void)setStart: (unsigned)theStart { start = theStart; }
-(unsigned)start { return start; }
-(void)setEnd: (unsigned)theEnd { end = theEnd; }
-(unsigned)theEnd { return end; }
-(void)setData: (void *)theData { data = theData; }
-(void)setSuperBlock: (Block *)theSuperBlock { superBlock = theSuperBlock; }
-(Block *)superBlock { return superBlock; }

-(char *)functionName
{
    return [blockGetter nameForPC: start];
}

-locals
{
    if (!locals)
	locals = [blockGetter blockValuesFor: data];
    return locals;
}

-(int)numLocals
{
    return [[self locals] count];
}

-(int)numSubValues
{
    if (superBlock)
	return [self numLocals] + [superBlock numSubValues];
    else
	return [self numLocals];
}

-(Value *)subValueAt: (int)nSubValue
{
    int numLocals = [self numLocals];
    
    if (nSubValue < numLocals)
	return [locals objectAt: nSubValue];
    else
	return [superBlock subValueAt: nSubValue - numLocals];
}

-(Value *)subValueAt: (int)nValue
           withFlags: (out unsigned *)pFlags
	     andName: (out char **)pName
{
    Value *value = [self subValueAt: nValue];

    *pFlags = [value flags];
    *pName  = [value valueName];
    return value;
}

-(Value *)subValueNamed: (char *)valueName index: (int *)pIndex
{
    int numSubValues = [self numSubValues], i;
    Value *subValue, *foundValue = nil;
    
    for (i = 0; !foundValue && (i < numSubValues); i++) {
	subValue = [self subValueAt: i];
	if (strcmp(valueName, [subValue valueName]) == 0)
	    foundValue = subValue;
    }
    if (foundValue)
	*pIndex = i - 1;
    return foundValue;
}

-(Value *)subValueFor: (Value *)oldValue
                 from: (int *)from
		index: (int *)index
{
    Value *value = [self subValueAt: *index];
    
    if (value != oldValue)
	value = [self subValueNamed: [oldValue valueName] index: index];
	
    if (value) {
	int isValid;
	
	*from = [value addressFrom: *from isValid: &isValid];
	if (!isValid)
	    value = nil;
    }
	
    return value;
}
	
@end

@implementation Frame

-(Value *)subValueAt:(int)nSubValue
{
    if (!block)
	block = [blockGetter makeBlock: pc];
    return [block subValueAt: nSubValue];
}

-(Value *)subValueNamed: (char *)valueName index: (int *)pIndex
{
    if (!block)
	block = [blockGetter makeBlock: pc];
    return [block subValueNamed: valueName index: pIndex];
}

-(Value *)subValueFor: (Value *)oldValue
                 from: (int *)from
		index: (int *)index
{
    if (!block)
	block = [blockGetter makeBlock: pc];
    return [block subValueFor: oldValue from: from index: index];
}

-(int)numSubValues
{
    if (!block)
	block = [blockGetter makeBlock: pc];
    return [block numSubValues];
}

-(int)hasSymbols
{
    if (block)
	return YES;
    else
	return [blockGetter hasSymbolsForPC: pc];
}

-(Value *)subValueAt: (int)nValue
           withFlags: (unsigned *)pFlags
	     andName: (char **)pName
{
    Value *value;
    
    if (!block)
	block = [blockGetter makeBlock: pc];
    value = [block subValueAt: nValue];
    *pFlags = [value flags];
    *pName  = [value valueName];
    return value;
}

-(char *)functionName
{
    return [blockGetter nameForPC: pc];
}

-(char *)functionNameAndHasSymbols: (int *)hasSymbols
{
    *hasSymbols = [self hasSymbols];
    return [self functionName];
}   

-(unsigned)frame { return frame; }
-(void)setFrame: (unsigned)theFrame { frame = theFrame; }
-(unsigned)pc { return pc; }
-(void)setPC: (unsigned)thePC { pc = thePC; }
@end

typedef struct _BlockID {
    @defs(Block)
} BlockID;

@implementation BlockList

-(int)indexForPC: (unsigned)pc
{
    BlockID **blockPtr, **bottom, **top, **center;
    
    if (!numElements)
	return -1;
    
    blockPtr = (BlockID **)dataPtr,
    bottom = blockPtr;
    top = bottom + numElements - 1;
    
    while (top > bottom) {
	center = bottom + (top - bottom) / 2;
	if ((*center)->start <= pc)
	    top = center;
	else
	    bottom = center + 1;
    }
    return top - blockPtr;
}
	    
-(Block *)blockForPC: (unsigned)pc
{
    BlockID **blockPtr, **next, **top;
    int index = [self indexForPC: pc];
    
    if (index == -1)
	return nil;
    
    blockPtr = (BlockID **)dataPtr;
    top = blockPtr + index;
    if (((*top)->start <= pc) && (pc < (*top)->end)) {
	for (next = top + 1;
	     (next < (blockPtr + numElements))
	     && (((*next)->start <= pc) && (pc < (*next)->end));
	     next++, top++);
	return (Block *)*top;
    } else
	return nil;
}

-addObject: anObject
{
    BlockID *block = (BlockID *)anObject, *blockPtr = (BlockID *)dataPtr,
            *top, *next;
    int index = [self indexForPC: block->start];
    
    if (index == -1)
	return [super addObject: anObject];
    
    top = blockPtr + index;
    if ((top->start <= block->start) && (block->end < top->end)) {
	for (next = top + 1;
	     (next < (blockPtr + numElements))
	     && ((next->start <= block->start) && (block->end < next->end));
	     index++, next++, top++);
    }
    [super insertObject: anObject at: index];
    return self;
}

@end