#import <objc/Object.h>
#import <objc/List.h>
#import <objc/maptable.h>
#import <mach/mach.h>
#import <remote/NXConnection.h>
#import <machkit/NXPort.h>
#import <appkit/Speaker.h>
#import <mach/cthreads.h>

#import "Value.h"
#import "Block.h"

@protocol GdbUser
-(oneway void)openFile: (char *)fileName onHost: (char *)hostName atTrueLine: (int)line;
-(oneway void)openFile: (char *)fileName onHost: (char *)hostName fromTrueLine: (int)startLine to: (int)endLine;
-(oneway void)setGdbName: (char *)name;
-(oneway void)setInterruptHandler: handler;
-(oneway void)disconnect;
-(void)frameChanged;
-(oneway void)programIsRunning;
-(void)programStopped;
-(void)programIsDead;
@end

@protocol GdbInternal
-(void)connectToSpeaker: (Speaker *)theSpeaker;
-(void)disconnect: (BOOL)tellClient;
-(void)openFile: (char *)fileName fromTrueLine: (int)firstLine to: (int)lastLine;
-(void)openFile: (char *)fileName atTrueLine: (int)line;
-(void)programStopped;
-(void)frameChanged;
-(oneway void)programIsRunning;
-(void)programIsDead;
@end

@protocol GdbControl
-(oneway void)breakInFile: (in char *)filename atLine: (int)line;
-(oneway void)untilInFile: (in char *)filename atLine: (int)line;
-(oneway void)step;
-(oneway void)next;
-(oneway void)continue;
-(oneway void)up;
-(oneway void)frame;
-(oneway void)down;
-(oneway void)finish;
-(oneway void)run;
-(oneway void)quit;
-(oneway void)printExpression: (in char *)selection;
-(oneway void)printExpressionIndirect: (in char *)selection;
-(oneway void)selectFrame: (int)frame;
-(oneway void)interrupt;
-(oneway void)setClient: (in id<GdbUser>)client;
-(int)getFramesUpTo: (int)maxFrames selFrame: (out int *)sel;
-(Frame *)frameAt: (int)frameIndex
     functionName: (out char **)pFunctionName
             from: (out unsigned int *)pFrom;
-(Frame *)frameAt: (int)frameIndex
             from: (out unsigned int *)pFrom;
-(char *)functionNameAt: (int)frameIndex;
-(char *)functionNameAt: (int)frameIndex hasSymbols: (int *)hasSymbols;
-(oneway void)errorOpeningFile: (char *)filename
                      fromLine: (int)start to: (int)end;

@end

@protocol Interrupt
-(oneway void)interrupt;
@end

@interface GdbController : Object
<GdbControl, GdbInternal, DataGetter, ErrorDelegate, TypeGetter, BlockGetter, 
 Interrupt, NXSenderIsInvalid>
{
    NXConnection *connection;
    NXConnection *interruptConnection;
    thread_t mainThread;
    NXMapTable *types;
    NXMapTable *values;
    NXMapTable *blocks;
    NXMapTable *demangledNames;
    Frame *frames;
    int numFrames;
    int numFramesAlloced;
    id <GdbUser>client;
    mutex_t clientLock;
    condition_t clientCond;
    BOOL lastToldAboutRunning;
    int maxFrames;
    int *selFrame;
}

-(Value *)makeValue: (struct symbol *)sym;
-(void)allocFrames: (int)nFrames;
@end