#import <objc/Object.h>
#import <objc/List.h>
#import <objc/HashTable.h>

#define ISPOINTER (1 << 0)
#define ISENUM    (1 << 1)
#define ISSCALER  (1 << 2)
#define ISFLOAT   (1 << 3)
#define ISDOUBLE  (1 << 4)
#define ISAGGREGATE (1 << 5)
#define ISCHAR	  (1 << 6)
#define ISARRAY   (1 << 7)
#define ISSIGNED  (1 << 8)
#define ISOBJECT  (1 << 9)
#define ISSTRING  (1 << 10)
#define ISFUNC  (1 << 11)

@protocol TypeGetter
-(Type *)makeType: (void *)data;
-(Type *)getTypeNamed: (char *)typeName from: (unsigned)from;
@end

@interface Type : Object
{
    char *typeName;
}
+(void)setTypeGetter: (id <TypeGetter>)theGetter;
+(id <TypeGetter>)typeGetter;
+(Type *)getTypeNamed: (char *)typeName from: (unsigned)from;
-(char *)typeName;
-(BOOL)isScaler;
-(BOOL)isEnum;
-(BOOL)isAggregate;
-(BOOL)isPointer;
-(BOOL)isFloat;
-(BOOL)isDouble;
-(BOOL)isString;
-(BOOL)isChar;
-(BOOL)isLazy;
-(BOOL)isArray;
-(BOOL)isSigned;
-(BOOL)isObject;
-(BOOL)isString;
-(BOOL)isFunc;
-(int)size;
-(unsigned)flags;
-(char *)getTypeName;
-(void)setTypeName: (char *)theName;
@end

@interface LazyType : Object
{
    void *data;
}
-(void)setData: (void *)theData;
-(void *)data;
-(Type *)type;
@end

@interface IndirectType : Object
{
    Type *type;
}
-(void)setType: (Type *)theType;
-(void)setLazyType: (void *)theData;
-(Type *)type;
@end

@interface AbsoluteType : Type
{
    int size;
}
-(void)setSize: (int)theSize;
-(int)size;
@end

@interface QualifiedType : IndirectType
@end

@interface Typedef : IndirectType
{
    char *typeName;
}
@end

@interface Function : Type
{
    Type *type;
}
-(void)setType: (Type *)theType;
-(void)setLazyType: (void *)theData;
-(Type *)type;
@end

@interface Volitile : QualifiedType
@end

@interface AggregateType : AbsoluteType
-(int)numSubValues;
-subValueAt: (int)nSubValue;
@end

@interface Void : AbsoluteType
@end

@interface ScalerType : AbsoluteType
@end

@interface Array : AggregateType
{
    Type *type;
    int number;
    HashTable *values;
    struct _aFlags {
	unsigned char isString : 1;
    } aFlags;
}
-(void)setType: (Type *)theType;
-(void)setLazyType: (void *)theData;
-(Type *)type;
-(void)setNumber: (int)theNumber;
-(int)number;
-(void)isString: (BOOL)isString;
@end

@interface UnionOrStructure : AggregateType
{
    List *fields;
}
-(void)setFields: (List *)theFields;
-(List *)fields;
@end

@interface Structure : UnionOrStructure
@end

@interface Union : UnionOrStructure
@end

@interface ArithmeticType : ScalerType
@end

@interface Pointer : ScalerType
{
    union _Indirect {
	Type *type;
	Value *value;
    } indirect;
    struct _pFlags {
	unsigned char isValue  : 1;
	unsigned char isString : 1;
	unsigned char isObject : 1;
	unsigned char isFunc   : 1;
    } pFlags;
    int indSize;
}
-(void)setType: (Type *)theType;
-(void)setLazyType: (void *)theData;
-(Type *)type;
-(void)isObject: (BOOL)isObject;
-(void)isString: (BOOL)isString;
-(void)isFunc:   (BOOL)isFunc;
-(void)setIndSize: (int)theSize;
-(int)indSize;
@end

@interface EnumVal : Object
{
    char *name;
    int val;
}
-(void)setName: (char *)theName;
-(char *)name;
-(void)setVal: (int)theVal;
-(int)val;
@end

@interface Enumeration : ScalerType
{
    List *vals;
}
-(void)setVals: (List *)theVals;
-(List *)vals;
-(EnumVal *)valFor: (int)scaler;
@end

@interface FloatingPoint : ArithmeticType
@end

@interface Integer : ArithmeticType
{
    BOOL isSigned;
}
-(void)setIsSigned: (BOOL)flag;
@end