#import "Value.h"

#import <objc/HashTable.h>
#import <stdlib.h>

static id <TypeGetter> typeGetter;

@implementation Type
+(void)setTypeGetter: (id <TypeGetter>)theGetter { typeGetter = theGetter; }
+(id <TypeGetter>)typeGetter { return typeGetter; }
+(Type *)getTypeNamed: (char *)name from: (unsigned)from
{ return [typeGetter getTypeNamed: name from: from]; }
-(BOOL)isScaler		{ return NO; }
-(BOOL)isEnum		{ return NO; }
-(BOOL)isAggregate	{ return NO; }
-(BOOL)isArray		{ return NO; }
-(BOOL)isPointer	{ return NO; }
-(BOOL)isFloat		{ return NO; }
-(BOOL)isDouble		{ return NO; }
-(BOOL)isChar		{ return NO; }
-(BOOL)isLazy		{ return NO; }
-(BOOL)isSigned		{ return NO; }
-(BOOL)isObject		{ return NO; }
-(BOOL)isString		{ return NO; }
-(BOOL)isFunc		{ return NO; }
-(int)size		{ return 0; }
-(unsigned)flags	{ return 0; }
-(char *)typeName
{
    if (!typeName)
	typeName = [self getTypeName];
    return typeName;
}
-(char *)getTypeName
{
    [self subclassResponsibility: _cmd];
    return NULL;
}
-(void)setTypeName: (char *)theName	{ typeName = theName; }
@end

@implementation LazyType
-(BOOL)isLazy		{ return YES; }
-(void)setData: (void *)theData		{ data = theData; }
-(void *)data		{ return data; }
-(Type *)type
{
    Type *type = [typeGetter makeType: data];
    [self free];
    return type;
}
@end

@implementation IndirectType
-(void)setType: (Type *)theType		{ type = theType; }

-(void)setLazyType: (void *)theData
{
    type = [LazyType new];
    [(LazyType *)type setData: theData];
}

-(Type *)type
{
    if ([type isLazy])
	type = [(LazyType *)type type];
    return type;
}

-forward:(SEL)aSelector :(marg_list)argFrame
{
    if ([[self type] respondsTo:aSelector])
        return [type performv: aSelector : argFrame];
    return [self doesNotRecognize: aSelector];
}

@end

@implementation Function
-(void)setType: (Type *)theType		{ type = theType; }

-(void)setLazyType: (void *)theData
{
    type = [LazyType new];
    [(LazyType *)type setData: theData];
}

-(Type *)type
{
    if ([type isLazy])
	type = [(LazyType *)type type];
    return type;
}

-forward:(SEL)aSelector :(marg_list)argFrame
{
    if ([[self type] respondsTo:aSelector])
        return [type performv: aSelector : argFrame];
    return [self doesNotRecognize: aSelector];
}

-(unsigned)flags { return ISFUNC; }
-(char *)getTypeName
{
    char *name = [[self type] typeName], *tName;
    tName = malloc(strlen(name) + 4);
    sprintf(tName, "%s ()", name);
    return tName;
}
-(BOOL)isFunc { return YES; }
@end

@implementation AbsoluteType
-(void)setSize: (int)theSize		{ size = theSize; }
-(int)size				{ return size; }
@end

@implementation QualifiedType
@end

@implementation Typedef
@end

@implementation Volitile
@end

@implementation AggregateType
-(BOOL)isAggregate	{ return YES; }
-(unsigned)flags	{ return ISAGGREGATE; }
-(int)numSubValues
{
    [self subclassResponsibility: _cmd];
    return 0;
}

-subValueAt: (int)nSubValue
{ 
    return [self subclassResponsibility: _cmd];
}

@end

@implementation Void
-(char *)getTypeName { return "void"; }
@end

@implementation ScalerType
-(unsigned)flags	{ return ISSCALER; }
-(BOOL)isScaler		{ return YES; }
@end

@implementation Array

-(BOOL)isArray	{ return YES; }
-(BOOL)isString
{
    return aFlags.isString ? YES : NO;
}

-(unsigned)flags { return ISAGGREGATE | ISARRAY | (aFlags.isString ? ISSTRING : 0); }

-(void)isString: (BOOL)isString
{
    aFlags.isString = isString;
}

static char *nameForNumber(int number)
{
    static HashTable *numbers;
    char *numberName, numberBuffer[100];
    int numberLength;
    static char *lowNumbers[] = {
	"[0]", "[1]", "[2]", "[3]", "[4]", "[5]", "[6]", "[7]", "[8]", "[9]",
	"[10]", "[11]", "[12]", "[13]", "[14]",
        "[15]", "[16]", "[17]", "[18]", "[19]"
    };
    
    if (number < (sizeof(lowNumbers) / sizeof(*lowNumbers)))
	return lowNumbers[number];
    if (!numbers)
	numbers = [HashTable newKeyDesc: "i" valueDesc: "*"];
    numberName = [numbers valueForKey: (void *)number];
    if (!numberName) {
	sprintf(numberBuffer, "[%d]", number);
	numberLength = strlen(numberBuffer) + 1;
	numberName = malloc(numberLength);
	memcpy(numberName, numberBuffer, numberLength);
	[numbers insertKey: (void *)number value: (void *)numberName];
    }
    return numberName;
}

-(void)setType: (Type *)theType		{ type = theType; }

-(void)setLazyType: (void *)theData
{
    type = [LazyType new];
    [(LazyType *)type setData: theData];
}

-(Type *)type
{
    if ([type isLazy])
	type = [(LazyType *)type type];
    return type;
}

-(void)setNumber: (int)theNumber	{ number = theNumber; }
-(int)number				{ return number; }
-(int)numSubValues			{ return number; }
-subValueAt: (int)nSubValue
{
    Value *value;
    if (!values)
	values = [HashTable newKeyDesc: "i"];
    value = [values valueForKey: (void *)nSubValue];
    if (!value) {
	value = [Value new];
	[value setValueName: nameForNumber(nSubValue)];
	[value setLocType: LTYPE_RELATIVE];
	[value setLocPlace: nSubValue * [[self type] size]];
	[value setType: type];
	[values insertKey: (void *)nSubValue value: value];
    }
    return value;
}

-(char *)getTypeName
{
    char *name = [[self type] typeName], num[20], *tName;
    sprintf(num, "%d", number);
    tName = malloc(strlen(name) + strlen(num) + 4);
    sprintf(tName, "%s [%s]", name, num);
    return tName;
}
@end

@implementation UnionOrStructure
-(void)setFields: (List *)theFields	{ fields = theFields; }
-(List *)fields				{ return fields; }
-(int)numSubValues			{ return [fields count]; }
-subValueAt: (int)nSubValue
{
    return [fields objectAt: nSubValue];
}
@end

@implementation Structure
@end

@implementation Union
@end

@implementation ArithmeticType
@end

@implementation Pointer
-(unsigned)flags	
{
    return ISPOINTER
         | ISSCALER 
	 | (pFlags.isString ? ISSTRING : 0)
	 | (pFlags.isObject ? ISOBJECT : 0)
	 | (pFlags.isFunc   ? ISFUNC   : 0);
}
-(BOOL)isPointer	{ return YES; }
-(void)setType: (Type *)theType		{ indirect.type = theType; }
-(void)setIndSize: (int)theSize { indSize = theSize; }
-(int)indSize { return indSize; }

-(void)setLazyType: (void *)theData
{
    indirect.type = [LazyType new];
    [(LazyType *)indirect.type setData: theData];
}

-(Type *)type
{
    if (pFlags.isValue)
	return [indirect.value type];
    else {
	if ([indirect.type isLazy])
	    indirect.type = [(LazyType *)indirect.type type];
	return indirect.type;
    }
}

-(BOOL)isString
{
    return pFlags.isString ? YES : NO;
}

-(BOOL)isObject
{
    return pFlags.isObject ? YES : NO;
}

-(BOOL)isFunc
{
    return pFlags.isFunc ? YES : NO;
}

-(void)isString: (BOOL)isString
{
    pFlags.isString = isString;
}

-(void)isObject: (BOOL)isObject
{
    pFlags.isObject = isObject;
}

-(void)isFunc: (BOOL)isFunc
{
    pFlags.isFunc = isFunc;
}

-(Value *)subValueAt: (int)nSubValue
{
    Type *type = [self type];
    
    if ([type isAggregate])
	return [(AggregateType *)type subValueAt: nSubValue];
    else if (nSubValue == 0) {
	if (!pFlags.isValue) {
	    Value *value = [Value new];
	    
	    [value setLocType: LTYPE_RELATIVE];
	    [value setLocPlace: 0];
	    [value setValueName: ""];
	    [value setType: type];
	    indirect.value = value;
	    pFlags.isValue = YES;
	}
	return indirect.value;
    } else
	return nil;
}

-(int)numSubValues
{
    Type *type = [self type];
    if ([type isAggregate])
	return [(AggregateType *)type numSubValues];
    else
	return 1;
}

-(char *)getTypeName
{
    char *name = [[self type] typeName], *tName;
    tName = malloc(strlen(name) + 3);
    sprintf(tName, "%s *", name);
    return tName;
}
@end

@implementation EnumVal
-(void)setName: (char *)theName		{ name = theName; }
-(char *)name				{ return name; }
-(void)setVal: (int)theVal		{ val = theVal; }
-(int)val				{ return val; }
@end

@implementation Enumeration
-(unsigned)flags		{ return ISSCALER | ISENUM; }
-(void)setVals: (List *)theVals		{ vals = theVals; }
-(List *)vals				{ return vals; }
-(BOOL)isEnum				{ return YES; }
-(BOOL)isScaler				{ return YES; }
-(EnumVal *)valFor: (int)scaler
{
    int i;
    EnumVal *enumVal, *foundVal;
    
    for (i = [vals count], foundVal = NULL; i-- && !foundVal;) {
	enumVal = [vals objectAt: i];
	if ([enumVal val] == scaler)
	    foundVal = enumVal;
    }
    return foundVal;
}
@end

@implementation FloatingPoint
-(unsigned)flags	{ return size == 4 ? ISFLOAT : ISDOUBLE; }
-(BOOL)isScaler		{ return NO; }
-(BOOL)isFloat		{ return size == 4; }
-(BOOL)isDouble		{ return size == 8; }
-(char *)getTypeName { return (size == sizeof(float)) ? "float" : "double"; }
@end

@implementation Integer
-(unsigned)flags
{
    unsigned int flags = ISSCALER;
    
    if (isSigned)
	flags |= ISSIGNED;
    if (size == 1)
	flags |= ISCHAR;
    return flags;
}
-(void)setIsSigned: (BOOL)flag		{ isSigned = flag; }
-(BOOL)isSigned				{ return isSigned; }
-(BOOL)isChar				{ return size == 1; }
-(char *)getTypeName
{
    const char *names[3][2] = {
	{"unsigned char", "char"},
	{"unsigned short", "short"},
	{"unsigned int", "int"}
    };
     
    return (char *)names[size == 1 ? 0 : size == 2 ? 2 : 2][isSigned];
}
@end