#import <objc/Object.h>

#import "ValueGetting.h"
#import "Type.h"

typedef enum _ValueError {
   ERROR_NEEDS_MORE = 1,
   ERROR_NOT_RELATIVE,
   ERROR_NOT_SCALER,
   ERROR_NOT_FLOAT,
   ERROR_NOT_ENUM,
   ERROR_NOT_AGGREGATE,
   ERROR_NOT_POINTER,
   ERROR_NOT_STRING,
   ERROR_NOT_FUNC,
   NUM_ERROR_CODES
} ValueError;

extern char *valueErrorStrings[NUM_ERROR_CODES];

typedef enum _LocType {
    LTYPE_ABSOLUTE,
    LTYPE_CONST,
    LTYPE_CONST_ADDRESS,
    LTYPE_MAX_INDEPENDENT = LTYPE_CONST_ADDRESS,
    LTYPE_RELATIVE,
    LTYPE_RELATIVE_ADDRESS,
    LTYPE_RELATIVE_BIT,
    LTYPE_REGISTER,
} LocType;

typedef union _Loc {
    int loc;
    void *address;
    int offset;
    unsigned int regNum;
    unsigned int constant;
    struct _LocBits {
	unsigned int bitPos : 27;
	unsigned int bitSize : 5;
    } locBits;
} LocPlace;

typedef struct _Location {
    LocType locType;
    LocPlace locPlace;
} Location;

@protocol DataGetter 
-(unsigned)getScalerValueFrom: (unsigned)from at: (Location *)location withSize: (int)size;
-(double)getDoubleValueFrom: (unsigned)from at: (Location *)location withSize: (int)size;
-(char *)getStringValueFrom: (unsigned)from at: (Location *)location;
-(char *)getStringValueForArrayFrom: (unsigned)from at: (Location *)location;
-(char *)getFunctionNameFrom: (unsigned)from at: (Location *)location;
-(unsigned)getAddressFrom: (unsigned)from at: (Location *)location;	
-(Type *)getObjectTypeFor: (unsigned)objectAddress;
-(int)isDataFrom: (unsigned)address for: (int)length;   
@end

@protocol ErrorDelegate
-valueError: (ValueError)valueError;
@end

@interface Value : Object
<ValueGetting>
{
    char *valueName;
    Type *_type;
    Location location;
}
+(void)setDataGetter: (id <DataGetter>)theGetter;
+(id <DataGetter>)dataGetter;
+(void)setErrorDelegate: (id <ErrorDelegate>)theGetter;
+(id <ErrorDelegate>)errorDelegate;
-(void)setValueName: (char *)theName;
-(void)setType: (Type *)theType;
-(void)setLocType: (LocType)locType;
-(LocType)locType;
-(void)setLocPlace: (int)locPlace;
-(int)locPlace;

@end