#import "gdb.h"

#include "voyeur.h"
#include "VoyeurSpeaker.h"

VoyeurSpeaker *voyeurSpeaker;
id voyeur;

void browseObject(char *exp, int from_tty)
{
    id object;
    int ok;
    int ret = 0;
    port_t voyeurPort, replyPort;
    if (!inferior_pid)
	error("Program must be running to browse objects\n");
    if (!voyeurSpeaker) {
	voyeurPort = NXPortFromName("AppInspector", NULL);
	if (voyeurPort) {
	    voyeurSpeaker = [VoyeurSpeaker new];
	    port_allocate(task_self(), &replyPort);
	    [voyeurSpeaker setReplyPort: replyPort];
	    [voyeurSpeaker setSendPort: voyeurPort];
	} else 
	    error("AppInspector isn't running.  Launch it and try again.");
    }
    if (voyeurSpeaker) {
	if (voyeur) 
	    ret = [voyeurSpeaker validateVoyeur: (int)voyeur ok: &ok];
	if ((!ret && !voyeur) || !ok)
	    ret = [voyeurSpeaker makeNewVoyeur: (int *)&voyeur
                                        forPid: inferior_pid];
	if (ret || !voyeur)
	    error("Couldn't create AppInspector for process.");
	if (!ret && exp && *exp) {
	    object = (id)parse_and_eval_address(exp);
	    ret = [voyeurSpeaker browseObject: (int)object
			             inVoyeur: (int)voyeur
			             withName: exp];
	}
	if (ret) {
	    voyeurSpeaker = nil;
	    error("Couldn't contact AppInspector.");
	}
    }
}

void
voyeurResume()
{
    if (voyeurSpeaker && voyeur)
	[voyeurSpeaker resumeApp: (int)voyeur];
}

void
_initialize_voyeur()
{
    add_com("browse", class_obscure, browseObject, "browse an object");
}
